////////////////////////////////////////////////////////////////////////////////
//
//  -------------------------------------
//  CleanMem.h - Memory Management System
//  -------------------------------------
//
//    Written / adapted by:
//      Tim Clegg, Oak Valley Hospital District
//      tclegg@ovhd.com
//    
//    Last Updated: 1-5-2004
//
//    Credit for this code also goes to the following sites / individuals:
//      http://www.flipcode.com/tutorials/tut_memleak.shtml
//      http://community.borland.com/article/0,1410,18041,00.html
//      http://www.experts-exchange.com/Programming/Programming_Languages/Cplusplus/Q_20806733.html
//
//    NOTE: concept and/or code snippets were taken from the any of the above
//      sites to create this code.  Therefore, I give credit where applicable to
//      the above authors, sites, etc.
//
//    Usage agreement:
//      By using this code, you agree to indemnify Oak Valley Hospital District,
//      Tim Clegg and all other contributors from any harm (physical, financial,
//      emotional, mental, socially, etc) that may result in your use of this
//      code.  Use this code at your own risk.
//
//    Other than that, enjoy the code!  If you find it useful, find bugs or an
//      upgrades to the code, let me know!
//
//    This code is free for use (commercial and non-commercial) as long as
//      this disclaimer stays intact within the code.
//
////////////////////////////////////////////////////////////////////////////////


/////////
//  --Instructions for use:--
//    1.  define any desired options
//    2.  include this file in your C++ file
//    3.  if you'd like to see a summary of memory leaks, run the function
//          BeginCheckingUnfreed(); where you'd like to start checking memory, and
//          DumpUnfreed(); at the point where you'd like to see memory leaks (at
//          the end of a program right before you exit / return is a good spot).
//
//
//  POSSIBLE DEFINABLE OPTIONS:
//  ---------------------------
//
// CLEAN_MEM_DISABLE
//
//  Defined:   All memory management is disabled.
//  Undefined: Memory management enabled.
//
// CLEAN_MEM_LOG_FILE
//
//  Defined:   Filename of log file that output is written to
//  Undefined: defaults to "leaks.log"
//
// CLEAN_MEM_NO_VERIFY
//
//  Defined:   Allocations and deallocations are not always verified.
//  Undefined: Extra verification of heap integrity on new and delete.
//
//
//  --End of instructions --
/////////

//
// Modified by Justin Pease for use with GameX
// (added strict memory allocation/deallocation verification
//  and more comprehensive information output)
//

#pragma once
#ifndef __CLEAN_MEM_H
  #define __CLEAN_MEM_H

	#include <stdio.h>
	#include <stdlib.h>
	#include <vector>
	#include <crtdbg.h>
	#include <assert.h>

// don't enable memory checking for code we're building in a static librarym
// otherwise we'd get 2 different delete functions in it and it wouldn't link
#ifdef _LIB
	#define CLEAN_MEM_DISABLE
#endif

#ifdef CLEAN_MEM_DISABLE

///	#if (defined(GAMEX_MAIN) && defined(_LIB)) || (defined(INCLUDING_FROM_MAIN_LOOP_SOURCE) && !defined(_LIB))
	#ifdef GAMEX_MAIN
		// if cleanMem was requested to be disabled,
		// do nothing but supply 2 empty functions:
		void BeginCheckingUnfreed(void) {}
		void DumpUnfreed(void) {}
	#else
		extern void BeginCheckingUnfreed(void);
		extern void DumpUnfreed(void);
	#endif

#else // !CLEAN_MEM_DISABLE:

///  #include <condefs.h>
//  #include <fstream.h>
//  #include <iostream.h>
//  #include <iomanip.h>

#pragma warning(disable: 4311)
#pragma warning(disable: 4267)
#pragma warning(disable: 4291)



	// make sure that there's a filename specified for the log file...
	#ifndef CLEAN_MEM_LOG_FILE
	#define CLEAN_MEM_LOG_FILE "leaks.log"
	#endif

///	#if (defined(GAMEX_MAIN) && defined(_LIB)) || (defined(INCLUDING_FROM_MAIN_LOOP_SOURCE) && !defined(_LIB))
	#ifdef GAMEX_MAIN

	// stuff for keeping track of allocations / deallocations
	typedef struct {
		DWORD address;
		DWORD size;
		const char * file;
		const char * function;
		DWORD line;
		bool isArray;
	} ALLOC_INFO;

	typedef std::vector<ALLOC_INFO*> AllocList;

	AllocList *allocList;

	char * allocDeleteFile;
	char * allocDeleteFunction;
	int allocDeleteLine = 0;

	bool leakDebugFileOpened = false;

	// outputs string to log file
	void OutputDebugString(char *debugStr)
	{
		// clear the output file if it hasn't been opened yet
		if(!leakDebugFileOpened)
		{
			FILE *out;
			if((out = fopen(CLEAN_MEM_LOG_FILE, "w")) != NULL)
				fclose(out);
			leakDebugFileOpened = true;
		}

		FILE *out;

		if((out = fopen(CLEAN_MEM_LOG_FILE, "a")) == NULL)
			fprintf(stderr, "Cannot open output file.\n");

		fputs(debugStr, out);

		fflush(out);
		fclose(out);
	};

	bool checkLeakScope = false;

	void BeginCheckingUnfreed(void)
	{
		checkLeakScope = true;
	}

	static void strcatAfter(char * str, const char * cat, int minFirstSegLen)
	{
		int len = strlen(str);

		if(len < minFirstSegLen)
		{
			for(int i = len ; i < minFirstSegLen ; i++)
				str[i] = ' ';

			str[minFirstSegLen] = '\0';
		}

		strcat(str, cat);
	}

	// shows a summary of unfreed memory
	void DumpUnfreed(void)
	{
		// don't dump anything if there were no memory leaks AND there is no existing dump file
		if(!allocList || allocList->empty())
		{
			FILE *test;
			if((test = fopen(CLEAN_MEM_LOG_FILE, "rb")) != NULL)
				fclose(test);
			else return;
		}

		AllocList::iterator i;
		DWORD totalSize = 0;
		char buf[1024];

		sprintf(buf, "\n-----------------------------------------------------------\n");			OutputDebugString(buf);
		sprintf(buf, " Begin Dump of Unfreed Memory:     (Output ID = %d)\n", (time(NULL)/4)%8092);	OutputDebugString(buf);
		sprintf(buf, "-----------------------------------------------------------\n\n");			OutputDebugString(buf);

		if(!allocList || allocList->empty())
		{
			sprintf(buf, "No memory leaks detected!\n");
			OutputDebugString(buf);
		}
		else
		{
			int maxLineNumLength = 0;
			int maxFileLength = 0;
			int maxFunctionLength = 0;

			// calculate maximum filename and function name lengths for a pretty printout
			for(unsigned int i = 0 ; i < allocList->size() ; i++)
			{
				ALLOC_INFO * info = (*allocList)[i];
				char fileStr [128];
				strncpy(fileStr, info->file, 128);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				int fileLength = strlen(fileStrPtr);
				if(maxFileLength < fileLength)
					maxFileLength = fileLength;

				int functionLength = strlen(info->function);
				if(maxFunctionLength < functionLength)
					maxFunctionLength = functionLength;

				char numStr [64];
				sprintf(numStr, "%d", info->line);

				int lineNumLength = strlen(numStr);
				if(maxLineNumLength < lineNumLength)
					maxLineNumLength = lineNumLength;
			}

			// loop through and print all relevant info about each memory leak in chronological order:
			for(unsigned int i = 0 ; i < allocList->size() ; i++)
			{
				ALLOC_INFO * info = (*allocList)[i];
				char fileStr [128];
				strncpy(fileStr, info->file, 128);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				bool alreadyDone = false;
				int bytes = 0;

				// for each allocList entry remaining (i.e. for each memory leak),
				for(unsigned int j = 0 ; j < allocList->size() ; j++)
				{
					ALLOC_INFO * info2 = (*allocList)[j];
					if(info->line == info2->line
					&& stricmp(info->file, info2->file) == 0) // if they're the same entries,
					{
						if(j < i)
						{
							alreadyDone = true; // already included this entry in a multi-entry output
							break;
						}
						else
						{
							bytes += info2->size; // increment number of entries this output corresponds with
						}
					}
				}

				if(alreadyDone)
					continue; // already included this entry in a multi-entry output

				char tempStr [256];

				sprintf(tempStr, "line %d ", info->line);
				strcpy(buf, tempStr);

				sprintf(tempStr, "of %s ", fileStrPtr);
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1);

				sprintf(tempStr, "in %s ", info->function);
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3+maxFileLength+1);

				sprintf(tempStr, "%d bytes\n", bytes);
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3+maxFileLength+1 + 3+maxFunctionLength+1);

				OutputDebugString(buf);
				totalSize += bytes;
			}
		}

		sprintf(buf, "\nTotal Unfreed: %d bytes\n\n", totalSize);						OutputDebugString(buf);
		sprintf(buf, "-----------------------------------------------------------\n");	OutputDebugString(buf);
		sprintf(buf, " End Dump of Unfreed Memory\n");									OutputDebugString(buf);
		sprintf(buf, "-----------------------------------------------------------\n");	OutputDebugString(buf);

		checkLeakScope = false;
	};

	// adds an allocated memory entry
	// TODO: if memory has already been allocated on the same line of the same file,
	//       this should just increase the 'size' value of the existing track
	//       as opposed to adding a whole new track and summing up their sizes later which is done now
	void AddTrack(DWORD addr,  DWORD asize,  const char *fname, const char *funcName, DWORD lnum, bool array)
	{
		ALLOC_INFO *info;

		if(!allocList) {
			allocList = new(AllocList);
		}

		info = new(ALLOC_INFO);
		info->address = addr;
		info->file = fname;
		info->function = funcName;
		info->line = lnum;
		info->size = asize;
		info->isArray = array;
		allocList->push_back(info);
	};

	// removes an allocated memory entry (when mem is de-allocated)
	void RemoveTrack(DWORD addr, bool array)
	{
		AllocList::iterator i;

		if(!allocList)
			return;

		for(i = allocList->begin(); i != allocList->end(); i++)
		{
			ALLOC_INFO *info = *i;

 			if(info->address == addr)
			{
				// if this assert fails, we mismatched new[] with delete, or new with delete[]
				assert(info->isArray == array);

				allocList->erase(i);
				return;
			}
		}

/*
		if(checkLeakScope)
		{
			char buf [1024];
			sprintf(buf, "Invalid Removal: address %d, via delete%s\n", addr, array?"[]":"");
			OutputDebugString(buf);
		}*/
	};

	void * __malloc(size_t size, const char *file, int line, const char *function)
	{
		void * ptr;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
		#endif

		ptr = (void *)malloc(size);

		if(checkLeakScope)
			AddTrack((DWORD)ptr, size, file, function, line, false);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(checkLeakScope && allocDeleteLine != 0)
			{
				char fileStr [128];
				strncpy(fileStr, file, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				char tempStr[2048];
				sprintf(tempStr, "NEW of %d bytes FAILED on line %d of %s\tin %s\n", size, line, fileStrPtr, function);
				OutputDebugString(tempStr);
///				DumpUnfreed();
///				#ifdef _DEBUG
///					DebugBreak(); // if it breaks here, check leaks.log and the rest of the call stack for more info
///					exit(-1);
///				#endif
				allocDeleteLine = 0;
			}
		}
		#endif

		return ptr;
	}

	void __free(void * mem)
	{
		if(!mem)
			return;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
			// If it breaks here, hit "Run" again then check the top of leaks.txt for more info
			// NOTE: a "user breakpoint" hit may be caused by forgetting to compile with Runtime = "Multi-threaded Debug DLL (/MDd)"
			if(checkLeakScope && _CrtIsValidHeapPointer(mem) == FALSE)
				throw "Invalid Pointer";

		#endif

		if(checkLeakScope)
			RemoveTrack((DWORD)mem, false);

		free(mem);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(checkLeakScope && allocDeleteLine != 0)
			{
				char fileStr [128];
				strncpy(fileStr, allocDeleteFile, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				char tempStr[2048];
				sprintf(tempStr, "DELETE of 0x%x FAILED on line %d of %s\tin %s\n", (DWORD)mem, allocDeleteLine, fileStrPtr, allocDeleteFunction);
				OutputDebugString(tempStr);
///				DumpUnfreed();
///				#ifdef _DEBUG
///					DebugBreak(); // if it breaks here, check leaks.log and maybe the rest of the call stack for more info
///					exit(-1);
///				#endif
				allocDeleteLine = 0;
			}
		}
		#endif
	}

	// overloaded new
	void * operator new(size_t size, const char *file, int line, const char *function)
	{
		return __malloc(size,file,line,function);
	};

	// overloaded new[]
	void * operator new[](size_t size, const char *file, int line, const char *function)
	{
		void * ptr;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
		#endif

		ptr = (void *)malloc(size);

		if(checkLeakScope)
			AddTrack((DWORD)ptr, size, file, function, line, true);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(checkLeakScope && allocDeleteLine != 0)
			{
				char fileStr [128];
				strncpy(fileStr, file, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				char tempStr[2048];
				sprintf(tempStr, "NEW[] of %d bytes FAILED on line %d of %s\tin %s\n", size, line, fileStrPtr, function);
				OutputDebugString(tempStr);
///				DumpUnfreed();
///				#ifdef _DEBUG
///					DebugBreak(); // if it breaks here, check leaks.log and maybe the rest of the call stack for more info
///					exit(-1);
///				#endif
				allocDeleteLine = 0;
			}
		}
		#endif

		return ptr;
	};

	// overloaded delete
	void operator delete(void * mem)
	{
		__free(mem);
	};

	// overloaded delete
	void operator delete[](void * mem)
	{
		if(!mem)
			return;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
			// If it breaks here, hit "Run" again then check the top of leaks.txt for more info
			// NOTE: a "user breakpoint" hit may be caused by forgetting to compile with Runtime = "Multi-threaded Debug DLL (/MDd)"
			if(checkLeakScope && _CrtIsValidHeapPointer(mem) == FALSE)
				throw "Invalid Pointer";

		#endif


		if(checkLeakScope)
			RemoveTrack((DWORD)mem, true);


		free(mem);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(checkLeakScope && allocDeleteLine != 0)
			{
				char fileStr [128];
				strncpy(fileStr, allocDeleteFile, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				char tempStr[2048];
				sprintf(tempStr, "DELETE[] of 0x%x FAILED on line %d of %s\tin %s\n", (DWORD)mem, allocDeleteLine, fileStrPtr, allocDeleteFunction);
				OutputDebugString(tempStr);
///				DumpUnfreed();
///				#ifdef _DEBUG
///					DebugBreak(); // if it breaks here, check leaks.log and the rest of the call stack for more info
///					exit(-1);
///				#endif
				allocDeleteLine = 0;
			}
///			Engine.ErrorDialog(tempStr);
		}
		#endif

	};

	#else

		struct ALLOC_INFO;
		typedef std::vector<ALLOC_INFO*> AllocList;
		extern AllocList *allocList;
		extern char * allocDeleteFile;
		extern char * allocDeleteFunction;
		extern int allocDeleteLine;
		extern void OutputDebugString(char *debugStr);
		extern void BeginCheckingUnfreed(void);
		extern void DumpUnfreed(void);
		extern void AddTrack(DWORD addr,  DWORD asize,  const char *fname, const char *funcName, DWORD lnum, bool array);
		extern void RemoveTrack(DWORD addr, bool array);
		extern void * operator new(size_t size, const char *file, int line, const char *function);
		extern void * operator new[](size_t size, const char *file, int line, const char *function);
		extern void operator delete(void * mem);
		extern void operator delete[](void * mem);
		extern void * __malloc(size_t size, const char *file, int line, const char *function);
		extern void __free(void * mem);
		extern bool checkLeakScope;
		extern bool leakDebugFileOpened;
	#endif

	// macro stuff for putting in file & line #

	#define new new(__FILE__, __LINE__, __FUNCTION__)

	#ifndef CLEAN_MEM_NO_VERIFY

		// for some reason delete cannot be overrided as easily as new can
		// so here's a tricky macro to pass the information anyway
		#define delete allocDeleteLine = __LINE__, allocDeleteFile = __FILE__, allocDeleteFunction = __FUNCTION__, delete

	#endif // !CLEAN_MEM_NO_VERIFY

	#define malloc(s) __malloc(s, __FILE__, __LINE__, __FUNCTION__)
	#define free(p) __free(p)

		
///	#define delete delete(__FILE__, __LINE__, __FUNCTION__)
/*
	#define DEBUG_NEW new (__FILE__, __LINE__, __FUNCTION__)
	#define new DEBUG_NEW
	#ifndef CLEAN_MEM_NO_VERIFY
		#define DEBUG_DELETE allocDeleteLine = __LINE__, strncpy(allocDeleteFile,__FILE__,127), strncpy(allocDeleteFunction,__FUNCTION__,127), delete
		#define delete DEBUG_DELETE
	#endif // !CLEAN_MEM_NO_VERIFY
*/
#endif // !CLEAN_MEM_DISABLE

#endif

